/*
    Akonadi OpenSync 0.22 implementation library
    Copyright (C) 2010 Tejas Guruswamy <masterpatricko@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

extern "C" {
    #include <opensync/opensync.h>
} //extern "C"

#include <QStringList>
#include <QCoreApplication>
#include <KComponentData>
#include <Akonadi/Control>
#include <Akonadi/CollectionFetchJob>
#include <Akonadi/CollectionFetchScope>
#include <Akonadi/ItemFetchJob>
#include <Akonadi/ItemCreateJob>
#include <Akonadi/ItemDeleteJob>
#include <Akonadi/ItemModifyJob>
#include <Akonadi/ItemFetchScope>
#include <akonadi/mimetypechecker.h>
#include <KCal/ICalFormat>
#include <KCal/Journal>
#include <KABC/VCardConverter>
#include "akonadisync.h"

AkonadiSync::AkonadiSync(qint64 contacts, qint64 events)
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    int argv = 0;
    char * argc = NULL;
    QCoreApplication* app = new QCoreApplication(argv, &argc);
    KComponentData* kcd = new KComponentData( "akonadi-sync" );
    
    Q_UNUSED( app );
    Q_UNUSED( kcd );
    
    osync_debug( "AKONADI-SYNC", 1, "Asked to use default collections contacts %lld and events %lld", 
                 contacts, events );
    m_contact_col = contacts;
    m_event_col = events;
	m_cols.clear();
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
}

bool AkonadiSync::initialize(int enabledtypes)
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    m_enabledtypes = enabledtypes;
    
	if ( !Akonadi::Control::start() ) {
        osync_debug( "AKONADI-SYNC", 1, "Couldn't start Akonadi!" );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
	}
	
	if ( !fetchCollections() ) {
        osync_debug( "AKONADI-SYNC", 1, "Couldn't find any collections!" );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
	}
	
	osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
	return true;
}

bool AkonadiSync::fetchCollections()
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    Akonadi::CollectionFetchJob *job = new Akonadi::CollectionFetchJob( Akonadi::Collection::root(),
                                                                        Akonadi::CollectionFetchJob::Recursive );
    QStringList mimeTypes;
    if (m_enabledtypes & ContactsEnabled)
        mimeTypes << "text/directory";
    if (m_enabledtypes & EventsEnabled)
        mimeTypes << "application/x-vnd.akonadi.calendar.event";
    if (m_enabledtypes & TodoEnabled)
        mimeTypes << "application/x-vnd.akonadi.calendar.todo";
    if (m_enabledtypes & NotesEnabled)
        mimeTypes << "application/x-vnd.kde.notes";
    
    job->fetchScope().setContentMimeTypes( mimeTypes );
	
	if ( !job->exec() ) {
        osync_debug( "AKONADI-SYNC", 1, "Collection fetch job failed! %s", job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
		return false;
    }
    
    m_cols = job->collections();
    osync_debug( "AKONADI-SYNC", 2, "Fetched %d akonadi collections", m_cols.count() );
	
	osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
	return true;
}

Akonadi::Collection AkonadiSync::getCollection(Akonadi::Collection::Id id)
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    if( !id ) {
        osync_debug( "AKONADI-SYNC", 1, "Returning empty collection" );
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return Akonadi::Collection();
    }
    
    //FIXME: Not returning a collection if specifying only the wanted collection!
    Akonadi::CollectionFetchJob *job = new Akonadi::CollectionFetchJob( Akonadi::Collection::root(), 
                                                                        Akonadi::CollectionFetchJob::Recursive );
    
    if ( !job->exec() ) {
        osync_debug( "AKONADI-SYNC", 1, "Job fetching collection %lld failed! %s", id, 
                     job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return Akonadi::Collection();
    }
    
    Akonadi::Collection::List cols = job->collections();
    
    if ( cols.isEmpty() ) {
        osync_debug( "AKONADI-SYNC", 1, "Job fetching collection %lld returned empty!", id );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return Akonadi::Collection();
    }
    
    Akonadi::Collection col;
    foreach ( Akonadi::Collection const &c, cols ) {
        if ( c.id() == id )
            col = c;
    }
    
    osync_debug( "AKONADI-SYNC", 2, "Fetched specific collection name %s id %lld", col.name().toAscii().constData(), col.id() );
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return col;
}

bool AkonadiSync::updateItem( OSyncChange* chg, bool getData )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    Akonadi::ItemFetchJob *job = new Akonadi::ItemFetchJob(     
        Akonadi::Item( QByteArray(osync_change_get_uid(chg)).toLongLong()) );
    if ( getData )
        job->fetchScope().fetchFullPayload();
    
    if ( !job->exec() ) {
        osync_debug( "AKONADI-SYNC", 1, "Job fetching item %s failed! %s", osync_change_get_uid(chg), 
                     job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    const Akonadi::Item::List items = job->items();
    
    if ( items.count() != 1 ) {
        osync_debug( "AKONADI-SYNC", 1, "Got %d items when looking for %s!", items.count(), osync_change_get_uid(chg) );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    Akonadi::Item item = items.first();
    
    osync_change_set_uid( chg, QByteArray::number( item.id() ));
    //When a new item is added, it has revision 0 then akonadi updates it to 1 causing a false change
    //hack to get around that by setting a min revision of 1
    osync_change_set_hash( chg, QByteArray::number( item.revision() > 0 ? item.revision() : 1 ));
    
    //Set the object format
    if ( Akonadi::MimeTypeChecker::isWantedItem(item, "text/directory")) {
        osync_change_set_objtype_string(chg, "contact");
        osync_change_set_objformat_string(chg, "vcard30");
    }
    if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.akonadi.calendar.event")) {
        osync_change_set_objtype_string(chg, "event");
        osync_change_set_objformat_string(chg, "vevent20");
    }
    if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.akonadi.calendar.todo" )) {
        osync_change_set_objtype_string(chg, "todo");
        osync_change_set_objformat_string(chg, "vtodo20");
    }
    if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.kde.notes" )) {
        osync_change_set_objtype_string(chg, "note");
        osync_change_set_objformat_string(chg, "vnote11");
    }
    
    QByteArray data;
    //The pointer to the data must last long enough to be passed to the other thread
    //QByteArray could be out of scope by that time; so use qstrdup instead
    if ( getData ) {
        data = getPayload(item);
        osync_change_set_data( chg, qstrdup(data), data.size(), TRUE );
    } else {
        osync_change_set_data( chg, qstrdup(data), data.size(), FALSE );
    }
    
    osync_debug( "AKONADI-SYNC", 2, "Answered request for item %lld revision %d", item.id(), item.revision() );
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return true;
}

QList<OSyncChange*> AkonadiSync::listItems()
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    QList<OSyncChange*> changes;
    foreach ( const Akonadi::Collection &col, m_cols  ) {
        osync_debug( "AKONADI-SYNC", 2, "Fetching items from collection id %lld name %s",
                     col.id(), col.name().toAscii().constData());
        Akonadi::ItemFetchJob *job = new Akonadi::ItemFetchJob( col );
        
        if ( !job->exec() ) {
            osync_debug( "AKONADI-SYNC", 1, "Job fetching items from collection id %lld failed! %s", 
                         col.id(), job->errorString().toAscii().constData() );
            osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
            return QList<OSyncChange*>();
        }
        
        const Akonadi::Item::List items = job->items();
        foreach ( const Akonadi::Item &item, items ) {
            osync_debug( "AKONADI-SYNC", 2, "Got item id %lld revision %d mime type %s", item.id(), item.revision(), item.mimeType().toAscii().constData() );
            
            Akonadi::MimeTypeChecker checker;
            if (m_enabledtypes & ContactsEnabled)
                checker.addWantedMimeType("text/directory");
            if (m_enabledtypes & EventsEnabled)
                checker.addWantedMimeType("application/x-vnd.akonadi.calendar.event");
            if (m_enabledtypes & TodoEnabled)
                checker.addWantedMimeType("application/x-vnd.akonadi.calendar.todo");
            if (m_enabledtypes & NotesEnabled)
                checker.addWantedMimeType("application/x-vnd.kde.notes");
            
            if (!checker.isWantedItem(item) )
                continue;

            //Make the new change to report
            OSyncChange* chg = osync_change_new();
            
            //Now set the uid of the object
            osync_change_set_uid( chg, QByteArray::number( item.id() ));
            
            //Set the object format
            if ( Akonadi::MimeTypeChecker::isWantedItem(item, "text/directory")) {
                osync_change_set_objtype_string(chg, "contact");
                osync_change_set_objformat_string(chg, "vcard30");
            }
            if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.akonadi.calendar.event")) {
                osync_change_set_objtype_string(chg, "event");
                osync_change_set_objformat_string(chg, "vevent20");
            }
            if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.akonadi.calendar.todo" )) {
                osync_change_set_objtype_string(chg, "todo");
                osync_change_set_objformat_string(chg, "vtodo20");
            }
            if ( Akonadi::MimeTypeChecker::isWantedItem(item, "application/x-vnd.kde.notes" )) {
                osync_change_set_objtype_string(chg, "note");
                osync_change_set_objformat_string(chg, "vnote11");
            }
            
            //Set the hash of the object (optional, only required if you use hashtables)
            //same hack as above
            osync_change_set_hash(chg, QByteArray::number( item.revision() > 0 ? item.revision() : 1 ));
            
            //Now you can set the data for the object
            //Set the last argument to FALSE if the real data
            //should be queried later in a "get_data" function
            char *data = NULL;
            osync_change_set_data(chg, data, sizeof(data), FALSE);
            
            changes.append(chg);
        }
    }
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return changes;
}

QByteArray AkonadiSync::getPayload( Akonadi::Item item )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    if ( item.mimeType() == "text/directory" ) {
        if ( item.hasPayload<KABC::Addressee>() ) {
            KABC::Addressee contact = item.payload<KABC::Addressee>();
            KABC::VCardConverter vcardconv;
            osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
            return vcardconv.createVCard(contact);
        }
    }
    else if ( item.mimeType() == "application/x-vnd.akonadi.calendar.event" ) {
        if ( item.hasPayload<KCal::Incidence::Ptr>() ) {
            KCal::Incidence::Ptr eventptr = item.payload<KCal::Incidence::Ptr>();
            KCal::ICalFormat icalconv;
            osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
            return icalconv.toString(eventptr.get()).toUtf8();
        }
    }
    else if ( item.mimeType() == "application/x-vnd.akonadi.calendar.todo" ) {
        if ( item.hasPayload<KCal::Incidence::Ptr>() ) {
            KCal::Incidence::Ptr eventptr = item.payload<KCal::Incidence::Ptr>();
            KCal::ICalFormat icalconv;
            osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
            return icalconv.toString(eventptr.get()).toUtf8();
        }
    }
    else if ( item.mimeType() == "application/x-vnd.kde.notes" ) {
        if ( item.hasPayload<KCal::Incidence::Ptr>() ) {
            KCal::Incidence::Ptr eventptr = item.payload<KCal::Incidence::Ptr>();
            QByteArray notestr;
            QTextStream out(&notestr, QIODevice::WriteOnly | QIODevice::Text );
            out << "BEGIN:VNOTE\n"
                << "VERSION:1.1\n"
                << "SUMMARY:" << eventptr->summary() << "\n"
                << "BODY:" << eventptr->description()  << "\n"
                << "DCREATED:" << eventptr->created().toString() << "\n"
                << "LAST-MODIFIED:" << eventptr->lastModified().toString() << "\n"
                << "END:VNOTE\n";
            out.flush();
            osync_debug("AKONADI-SYNC", 1, "Note created with vnote text: %s\n", notestr.constData());
            osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
            return notestr;
        }
    }
    
    osync_debug( "AKONADI-SYNC", 1, "Item id %lld mime type %s doesn't have expected payload!", 
                 item.id(), item.mimeType().toAscii().constData() );
    osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
    return QByteArray();
}

void AkonadiSync::setPayload( Akonadi::Item& item, OSyncChange* chg )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    if( !osync_change_has_data(chg) ) {
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        osync_debug( "AKONADI-SYNC", 1, "No data provided to set payload!" );
        return;
    }
    
    QByteArray data = osync_change_get_data(chg);
    osync_trace( TRACE_SENSITIVE, "Creating item containing data %s", data.constData() );
    
    OSyncObjType* otype = osync_change_get_objtype(chg);
    QByteArray type = osync_objtype_get_name(otype);
    
    if ( type == "contact" ) {
        KABC::VCardConverter vcardconv;
        KABC::Addressee contact = vcardconv.parseVCard(data);
        item.setMimeType( "text/directory" );
        item.setPayload<KABC::Addressee>(contact);
    } else if ( type == "event" ) {
        KCal::ICalFormat icalconv;
        KCal::Incidence::Ptr event = KCal::Incidence::Ptr( icalconv.fromString(data) );
        item.setMimeType( "application/x-vnd.akonadi.calendar.event" );
        item.setPayload<KCal::Incidence::Ptr>(event);
    } else if ( type == "todo" ) {
        KCal::ICalFormat icalconv;
        KCal::Incidence::Ptr event = KCal::Incidence::Ptr( icalconv.fromString(data) );
        item.setMimeType( "application/x-vnd.akonadi.calendar.todo" );
        item.setPayload<KCal::Incidence::Ptr>(event);
    } else if ( type == "note" ) {
        QString summary;
        QString body;
        QString dateCreated;
        QString lastModified;
        QTextStream in( &data, QIODevice::ReadOnly | QIODevice::Text );
        while ( !in.atEnd() ) {
            QString str = in.readLine();
            QString key = str.split(":").first();
            QStringList value = str.split(":").mid(1);
            if (key == "SUMMARY")
                summary = value.join(":");
            else if (key == "BODY")
                body = value.join(":");
            else if (key == "DCREATED")
                dateCreated = value.join(":");
            else if (key == "LAST-MODIFIED")
                lastModified = value.join(":");
            else
                continue;
        }
        KCal::Journal::Ptr event = KCal::Journal::Ptr( new KCal::Journal() );
        event->setSummary(summary);
        event->setDescription(body);
        if (!lastModified.isEmpty())
            event->setLastModified(KDateTime::fromString(lastModified));
        if (!dateCreated.isEmpty())
            event->setCreated(KDateTime::fromString(dateCreated));
        item.setMimeType( "application/x-vnd.kde.notes" );
        item.setPayload<KCal::Journal::Ptr>(event);
    }
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
}

bool AkonadiSync::addItem ( OSyncChange* chg )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    OSyncObjType* otype = osync_change_get_objtype(chg);
    QByteArray type = osync_objtype_get_name(otype);
    QString mimeType;
    Akonadi::Collection recepient;
    if ( type == "contact" ) {
        mimeType = "text/directory";
        recepient = getCollection(m_contact_col);
    } else if ( type == "event" ) {
        mimeType = "application/x-vnd.akonadi.calendar.event";
        recepient = getCollection(m_event_col);
    } else if ( type == "todo" ) {
        mimeType = "application/x-vnd.akonadi.calendar.todo";
        recepient = getCollection(m_event_col);
    } else if ( type == "note" ) {
        mimeType = "application/x-vnd.kde.notes";
    }
    
    if ( !recepient.isValid() ) {
        foreach( Akonadi::Collection col, m_cols ) {
            if( Akonadi::MimeTypeChecker::isWantedCollection(col, mimeType) )
                recepient = col;
        }
        osync_debug( "AKONADI-SYNC", 1, "No configured collection, using id %lld for mime type %s instead", 
                     recepient.id(), recepient.mimeType().toAscii().constData() );
    }
    
    Akonadi::Item item;
    setPayload(item, chg);
    
    Akonadi::ItemCreateJob *job = new Akonadi::ItemCreateJob( item, recepient );
    if ( !job->exec() ){
        osync_debug( "AKONADI-SYNC", 1, "Job creating item id %lld in collection id %lld failed! %s", 
                     item.id(), recepient.id(), job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    //Update the uid so we get the right item when fetching
    item = job->item();
    osync_change_set_uid( chg, QByteArray::number( item.id() ));
    osync_debug( "AKONADI-SYNC", 2, "Added item id %lld revision %d", item.id(), item.revision() );
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return true;
}

bool AkonadiSync::deleteItem ( OSyncChange* chg )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    Akonadi::Item item( QByteArray(osync_change_get_uid(chg)).toLongLong() );
    Akonadi::ItemDeleteJob *job = new Akonadi::ItemDeleteJob( item );
    if ( !job->exec() ){
        osync_debug( "AKONADI-SYNC", 1, "Job deleting item id %lld failed! %s", 
                     item.id(), job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    osync_debug( "AKONADI-SYNC", 2, "Deleted item %lld revision %d", item.id(), item.revision() );
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return true;
}

bool AkonadiSync::modifyItem ( OSyncChange* chg )
{
    osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
    
    Akonadi::ItemFetchJob *fetchJob = new Akonadi::ItemFetchJob( 
            Akonadi::Item( QByteArray(osync_change_get_uid(chg)).toLongLong() ));
    fetchJob->fetchScope().fetchFullPayload();
    
    if ( !fetchJob->exec() ) {
        osync_debug( "AKONADI-SYNC", 1, "Job fetching item id %s for modification failed! %s", 
                     osync_change_get_uid(chg), fetchJob->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    const Akonadi::Item::List items = fetchJob->items();
    
    if ( items.count() == 0 ) {
        //Item doesn't exist, turn this into an addition
        osync_debug( "AKONADI-SYNC", 2, "Was asked to modify item %s, doesn't exist. Will add instead.", 
                     osync_change_get_uid(chg) );
        osync_trace(TRACE_INTERNAL, "%s", __PRETTY_FUNCTION__);
        return addItem(chg);
    } else if ( items.count() > 1 ) {
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    Akonadi::Item item = items.first();
//    item.clearPayload();
    setPayload( item, chg );
    Akonadi::ItemModifyJob *job = new Akonadi::ItemModifyJob( item );
    job->setIgnorePayload(false);
    if ( !job->exec() ){
        osync_debug( "AKONADI-SYNC", 1, "Job modifying item id %lld failed! %s", 
                     item.id(), job->errorString().toAscii().constData() );
        osync_trace(TRACE_EXIT_ERROR, "%s", __PRETTY_FUNCTION__);
        return false;
    }
    
    //Update the uid so we get the right item when fetching
    item = job->item();
    osync_change_set_uid( chg, QByteArray::number( item.id() ));
    osync_debug( "AKONADI-SYNC", 2, "Modified item %lld revision %d", item.id(), item.revision() );
    
    osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    return true;
}

extern  "C"
{
    KDE_EXPORT AkonadiSyncWrapper* new_AkonadiSync(qint64 contacts, qint64 events) {
        AkonadiSyncWrapper* imp = new AkonadiSync( contacts, events );
        return imp;
    }
}

#include <akonadisync.moc>
